package restclient;

import javax.json.*;
import javax.mail.MessagingException;
import java.io.*;
import java.text.DecimalFormat;
import java.util.*;

/**
 * JSON-REST-Client zur Berechnung des RCI (Rules Compliance Index) aller Projekte fuer SonarQube,
 * getestet mit den SonarQube-Versionen 6.3 und 7.0.
 * {@link "https://docs.sonarqube.org/display/DEV/Web+API"}
 * {@link "http://localhost:9000/web_api"}
 * {@link "http://localhost:9000/api/components/search?qualifiers=TRK"}
 * {@link "http://localhost:9000/api/measures/component?metricKeys=ncloc&componentKey={projName}"}
 * {@link "http://localhost:9000/api/issues/search?statuses=OPEN&severities=BLOCKER&facets=severities&facetMode=count&projectKeys={projName}"}
 */
public class SonarQubeRci
{
   private static final double RCI_LIMIT          = 95.1;
   private static final double CRIT_PR_LIMIT      = 0.01;
   private static final double MAJR_PR_LIMIT_HTML = 0.80;
   private static final double MAJR_PR_LIMIT_MAIL = 0.90;

   public static void main( String[] args ) throws MessagingException
   {
      System.out.println(
            "\n\n---------------------------------------------------------------------------------------------------------------\n" +
            "Berechnung des RCI (Rules Compliance Index) aller Projekte fuer SonarQube 6.3 bis 7.0,\n" +
            "1. Parameter: SonarQube-Basis-URL, 2. Parameter: Severity-RCI-Wichtung,\n" +
            "3. Parameter: E-Mail-Empfaenger, 4. Parameter: E-Mail-SMTP-Host, z.B.:\n" +
            "  java -cp target\\RestClient-jar-with-dependencies.jar restclient.SonarQubeRci http://localhost:9000 10/5/3/1/0\n" +
            "Der E-Mail-Empfaenger kann auch per Umgebungsvariable 'SonarQubeRciEMailEmpfaenger' gesetzt werden.\n" +
            "---------------------------------------------------------------------------------------------------------------\n" );

      String basisUrl       = ( args != null && args.length > 0 && args[0].length() > 0 ) ? args[0] : "http://localhost:9000";
      String rciWeightsStr  = ( args != null && args.length > 1 && args[1].length() > 0 ) ? args[1] : "10/5/3/1/0";
      String mailEmpfaenger = ( args != null && args.length > 2 && args[2].length() > 0 ) ? args[2] : null;
      String mailSmtpHost   = ( args != null && args.length > 3 && args[3].length() > 0 ) ? args[3] : null;

      if( mailEmpfaenger == null || mailEmpfaenger.isEmpty() ) {
         mailEmpfaenger = System.getenv( "SonarQubeRciEMailEmpfaenger" );
      }

      int[] rciWeights = new int[] { 10, 5, 3, 1, 0 };
      String[] ss = rciWeightsStr.replace( '/', ',' ).split( "," );
      if( ss.length == rciWeights.length ) {
         for( int i = 0; i < rciWeights.length; i++ ) {
            rciWeights[i] = Integer.parseInt( ss[i] );
         }
      }

      System.out.println( "SonarQube-URL:         " + basisUrl + "/projects" );
      System.out.println( "Severity-RCI-Wichtung: " + rciWeights[0] + "/" + rciWeights[1] + "/" + rciWeights[2] + "/" +
                                                      rciWeights[3] + "/" + rciWeights[4] + " (Blocker/Critical/Major/Minor/Info)" );
      if( mailEmpfaenger != null && !mailEmpfaenger.isEmpty() ) {
         System.out.println( "E-Mail-Empfaenger:     " + mailEmpfaenger );
      }

      List<SonarQubeProjektInfos> projekteInfos = findeProjekteNamen( basisUrl );
      ermittleLinesOfCodeAnzahl( basisUrl, projekteInfos );
      ermittleViolations( basisUrl, rciWeights, projekteInfos );

      System.out.println( "Anzahl der Projekte:   " + projekteInfos.size() );
      SonarQubeProjektInfos summenInfos = berechneSummarischeErgebnisse( projekteInfos, rciWeights );
      String summarischesErgebnis = "Summarisches Ergebnis: " + "LOC: " + summenInfos.linesOfCode + ", Blocker: " + summenInfos.blocker +
            ", Crit.: " + summenInfos.critical + ", Major: " + summenInfos.major + ", Minor: " + summenInfos.minor + ", RCI: ";
      summarischesErgebnis = summarischesErgebnis + (( summenInfos.rci <= 96.0 && summenInfos.rci > 95.9 )
            ? (summenInfos.rci + "\n") : String.format( "%.1f\n", summenInfos.rci ));
      System.out.println( summarischesErgebnis );

      ProjektInfosComparator comp = new ProjektInfosComparator();
      comp.sortcolumn = "projName1";
      projekteInfos.sort( comp );
      speichereInCsv( projekteInfos, "SonarQube-RCI.csv" );
      speichereInHtml( projekteInfos, "SonarQube-RCI-sortName.html" );
      comp.sortcolumn = "rci";
      projekteInfos.sort( comp );
      speichereInHtml( projekteInfos, "SonarQube-RCI-sortRci.html" );

      List<SonarQubeProjektInfos> projekteInfosKritisch = findeKritischeProjekte( projekteInfos );
      System.out.println( ausgabeInString( "Kritische Projekte (in der HTML-Datei sind kritische Werte rot):", projekteInfosKritisch ) );

      if( summenInfos.blocker > 0 || summenInfos.critical > 100 || summenInfos.rci <= 96.0 ||
            (projekteInfosKritisch != null && !projekteInfosKritisch.isEmpty()) ) {
         String inhalt =
               "SonarQube-URL:         " + basisUrl + "/projects\n" +
                     "Anzahl der Projekte:   " + projekteInfos.size() + "\n" + summarischesErgebnis + "\n" +
                     ausgabeInString( "Kritische Projekte (in der HTML-Datei sind kritische Werte rot):", projekteInfosKritisch ) +
                     "\nSiehe: SonarQube-RCI-sortRci.html";
         EMailVersendenUtil.versendeEMail(
               "SonarQube: Zu viele Violations", inhalt, "SonarQube@meinprojekt.meinefirma.de",
               mailEmpfaenger, mailSmtpHost );
      }
   }

   private static List<SonarQubeProjektInfos> findeProjekteNamen( String basisUrl )
   {
      List<SonarQubeProjektInfos> projekteInfos = new ArrayList<>();
      JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl( basisUrl + "/api/components/search?qualifiers=TRK" );
      JsonArray components = jsonObj.getJsonArray( "components" );
      for( JsonObject comp : components.getValuesAs( JsonObject.class ) ) {
         SonarQubeProjektInfos pi = new SonarQubeProjektInfos();
         pi.projName2 = comp.getString( "key" );
         pi.projName1 = pi.projName2;
         int i = pi.projName1.indexOf( ':' );
         if( i > 0 ) {
            pi.projName1 = ( pi.projName1.startsWith( "schnittstellen" ) ) ? pi.projName1.substring( 0, i ) : pi.projName1.substring( ++i );
         }
         projekteInfos.add( pi );
      }
      return projekteInfos;
   }

   private static void ermittleLinesOfCodeAnzahl( String basisUrl, List<SonarQubeProjektInfos> projekteInfos )
   {
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl(
               basisUrl + "/api/measures/component?metricKeys=ncloc&componentKey=" + pi.projName2 );
         JsonArray measures = jsonObj.getJsonObject( "component" ).getJsonArray( "measures" );
         for( JsonObject m : measures.getValuesAs( JsonObject.class ) ) {
            if( "ncloc".equalsIgnoreCase( m.getString( "metric" ) ) ) {
               try {
                  pi.linesOfCode += Integer.parseInt( m.getString( "value" ) );
               } catch( NumberFormatException e ) {
                  System.out.println( "Lines of Code konnte nicht geparsed werden aus: " + m.getString( "value" ) );
               }
            }
         }
      }
   }

   private static void ermittleViolations( String basisUrl, int[] rciWeights, List<SonarQubeProjektInfos> projekteInfos )
   {
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         JsonObject jsonObj = JsonObjectFromUrlUtil.getJsonObjectFromUrl( basisUrl +
                     "/api/issues/search?statuses=OPEN&severities=BLOCKER&facets=severities&facetMode=count&projectKeys=" +
                     pi.projName2 );
         JsonArray facets = jsonObj.getJsonArray( "facets" );
         for( JsonObject violationFacets : facets.getValuesAs( JsonObject.class ) ) {
            JsonArray violationValues = violationFacets.getJsonArray( "values" );
            for( JsonObject viol : violationValues.getValuesAs( JsonObject.class ) ) {
               String val = viol.getString( "val" );
               int    cnt = viol.getInt( "count" );
               if( "BLOCKER".equalsIgnoreCase(  val ) ) { pi.blocker  = cnt; }
               if( "CRITICAL".equalsIgnoreCase( val ) ) { pi.critical = cnt; }
               if( "MAJOR".equalsIgnoreCase(    val ) ) { pi.major    = cnt; }
               if( "MINOR".equalsIgnoreCase(    val ) ) { pi.minor    = cnt; }
               if( "INFO".equalsIgnoreCase(     val ) ) { pi.info     = cnt; }
               pi.rci = berechneRci(  pi.linesOfCode, rciWeights,  pi.blocker,  pi.critical,  pi.major,  pi.minor,  pi.info );
            }
         }
      }
   }

   private static SonarQubeProjektInfos berechneSummarischeErgebnisse( List<SonarQubeProjektInfos> projekteInfos, int[] rciWeights )
   {
      SonarQubeProjektInfos summenInfos = new SonarQubeProjektInfos();
      int n = projekteInfos.size();
      if( n > 0 ) {
         for( SonarQubeProjektInfos pi : projekteInfos ) {
            summenInfos.linesOfCode += pi.linesOfCode;
            summenInfos.blocker     += pi.blocker;
            summenInfos.critical    += pi.critical;
            summenInfos.major       += pi.major;
            summenInfos.minor       += pi.minor;
            summenInfos.rci         += pi.rci;
         }

         int issueWeight = rciWeights[0] * summenInfos.blocker + rciWeights[1] * summenInfos.critical +
                           rciWeights[2] * summenInfos.major + rciWeights[3] * summenInfos.minor + rciWeights[4] * summenInfos.info;
         summenInfos.rci = Math.max( 100. - ((double) issueWeight / summenInfos.linesOfCode) * 100., 0 );
      }
      return summenInfos;
   }

   private static List<SonarQubeProjektInfos> findeKritischeProjekte( List<SonarQubeProjektInfos> projekteInfos )
   {
      List<SonarQubeProjektInfos> projekteInfosKritisch = new ArrayList<>();
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         if( (pi.rci > 0 && pi.rci <= RCI_LIMIT) || pi.blocker > 0 ||
               100. * pi.critical / pi.linesOfCode >= CRIT_PR_LIMIT ||
               100. * pi.major    / pi.linesOfCode >= MAJR_PR_LIMIT_MAIL ) {
            projekteInfosKritisch.add( pi );
         }
      }
      return projekteInfosKritisch;
   }

   private static String ausgabeInString( String titelZeile, List<SonarQubeProjektInfos> projekteInfos )
   {
      if( projekteInfos == null || projekteInfos.isEmpty() ) { return ""; }
      int len = 0;
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         if( pi.projName1 != null && pi.projName1.length() > len ) { len = pi.projName1.length(); }
      }
      StringBuilder sb = new StringBuilder( titelZeile + "\n" );
      for( SonarQubeProjektInfos pi : projekteInfos ) {
         sb.append( String.format( "%-" + (len + 1) + "s Loc: %5d, Blocker: "  + pi.blocker +
               ", Critical: " + pi.critical + ", Major: %3d, Minor: %3d, Info: %3d, --> RCI: %.1f\n",
               pi.projName1 + ":", pi.linesOfCode, pi.major, pi.minor, pi.info, pi.rci ) );
      }
      return sb.toString();
   }

   private static void speichereInHtml( List<SonarQubeProjektInfos> projekteInfos, String htmlFilename )
   {
      try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( htmlFilename ), "UTF-8") ) ) {
         out.write( "<html><body><table border=0 cellspacing=8 cellpadding=0>\n" );
         out.write( "<tr bgcolor='#E7EAEA'><th>Projekt</th><th>RCI</th><th>LOC</th>" +
               "<th>Blocker</th><th colspan='2'>Critical</th><th colspan='2'>Major</th><th>Minor</th><th>Info</th></tr>\n" );
         for( SonarQubeProjektInfos pi : projekteInfos ) {
            out.write( "<tr><td>" + pi.projName1 + "</td>" +
                  "<td align='right'>" + checkRciLimitForHtml( pi.rci, RCI_LIMIT ) + "</td>" +
                  "<td align='right'>" + pi.linesOfCode + "</td>" +
                  "<td align='center'>" + checkBlockerForHtml( pi.blocker ) + "</td>" +
                  "<td align='right'>" + pi.critical + "</td><td>" + calcProzentForHtml( pi.critical, pi.linesOfCode, CRIT_PR_LIMIT ) + "</td>" +
                  "<td align='right'>" + pi.major    + "</td><td>" + calcProzentForHtml( pi.major,    pi.linesOfCode, MAJR_PR_LIMIT_HTML ) + "</td>" +
                  "<td align='right'>" + pi.minor + "</td><td align='right'>" + pi.info + "</td></tr>\n" );
         }
         out.write( "</table></body></html>\n" );
      } catch( IOException e ) {
         e.printStackTrace();
      }
   }

   private static String checkBlockerForHtml( int blocker )
   {
      String blockerStr = "" + blocker;
      return ( blocker > 0 ) ? "<font color='red'><b>" + blockerStr + "</b></font>" : blockerStr;
   }

   private static String checkRciLimitForHtml( double rci, double limit )
   {
      String rciStr = String.format( "%.1f", rci );
      return ( rci > 0. && rci <= limit ) ? "<font color='red'><b>" + rciStr + "</b></font>" : rciStr;
   }

   private static String calcProzentForHtml( int violations, int linesOfCode, double limit )
   {
      if( violations == 0 || linesOfCode == 0 ) { return ""; }
      double prozent = 100. * violations / linesOfCode;
      String s = String.format( "<small>(%1$,.2f%%)</small>&nbsp;", prozent );
      return ( prozent >= limit ) ? "<font color='red'><b>" + s + "</b></font>" : s;
   }

   private static void speichereInCsv( List<SonarQubeProjektInfos> projekteInfos, String csvFilename )
   {
      try( BufferedWriter out = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( csvFilename ), "UTF-8") ) ) {
         out.write( "Projekt;RCI;LOC;Blocker;Critical;Major;Minor;Info;\n" );
         for( SonarQubeProjektInfos pi : projekteInfos ) {
            DecimalFormat df = (DecimalFormat) DecimalFormat.getInstance( Locale.US );
            df.applyPattern( "0.0" );
            String rci = df.format( pi.rci );
            out.write( pi.projName1 + ";" + rci + ";" + pi.linesOfCode + ";" +
                  pi.blocker + ";" + pi.critical + ";" + pi.major + ";" + pi.minor + ";" + pi.info + ";\n" );
         }
      } catch( IOException e ) {
         e.printStackTrace();
      }
   }

   private static double berechneRci( int linesOfCode, int[] rciWeights, int blocker, int critical, int major, int minor, int info )
   {
      if( linesOfCode <= 0 ) { return 0.0; }
      int issueWeight = rciWeights[0] * blocker + rciWeights[1] * critical + rciWeights[2] * major + rciWeights[3] * minor + rciWeights[4] * info;
      return Math.max( 100. - ((double) issueWeight / linesOfCode) * 100., 0 );
   }

   static class SonarQubeProjektInfos
   {
      String projName1;
      String projName2;
      int linesOfCode;
      int blocker;
      int critical;
      int major;
      int minor;
      int info;
      double rci;
   }

   static class ProjektInfosComparator implements Comparator<SonarQubeProjektInfos>
   {
      String sortcolumn;

      @Override
      public int compare( SonarQubeProjektInfos pi1, SonarQubeProjektInfos pi2 ) {
         if( "rci".equalsIgnoreCase( sortcolumn ) ) {
            return Double.compare( pi1.rci, pi2.rci );
         } else {
            return pi1.projName1.compareTo( pi2.projName1 );
         }
      }
   }
}
